/* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
 * ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
 * PARTICULAR PURPOSE.
 *
 * Copyright (C) 1994  Microsoft Corporation.  All Rights Reserved.
 *
 *  MODULE:   compq.c
 *
 *  PURPOSE:
 *	  handles input and output queues for the CompBunch function
 *
 *  FUNCTIONS:
 *	  CompBunchPass(void)
 *	  RPCCompBunchRequest(<bunch_data>)
 *	  RPCGetCompBunchResult(<bunch_data>)
 *	  RPCCompBunchReqAccept(handle_t hHandle)
 *	  RPCCompBunchRetAvail(handle_t hHandle, LONG lBunchId)
 *	  InitCompQueues(void)
 *	  QuickDispBunch(LONG lBunchId, LONG lBunchLen, LONG alIndex[])
 *	  DispCompBunchOutQ(TCHAR *szPrefix)
 *	  DispCompBunchInQ(TCHAR *szPrefix)
 *
 *  COMMENTS:
 *    The functions beginning with 'RPC' are client-server interfaces defined
 *    in the MIDL file.
 */

#include	"dcomp.h"
#include	"dcomp_s.h"

void		DispCompBunchOutQ(TCHAR *szPrefix);
void		DispCompBunchInQ(TCHAR *szPrefix);
void		QuickDispBunch(LONG lBunchId, LONG lBunchLen, LONG alIndex[]);


//
//  FUNCTION: CompBunchPass
//
//  PURPOSE:
//    compress a bunch from the input queue and place it on the output queue
//
//  PARAMETERS:
//
//  RETURN VALUE:
//
//  COMMENTS:
//    The compression component of a single pass of CodecLoop.  Returns
//    immediately if there are no bunches to process.  If an attempt to
//    compress a bunch fails, the bunch is re-queued for a re-try on the
//    next pass.
//

void		CompBunchPass(void)
{
	LONG				nRequ;
	LONG				lBunchLen;
	LPBITMAPINFOHEADER	lpbihRec;
	LPBITMAPINFOHEADER	lpbihUnc;


	// Check there is a return slot free

	if (ReserveSlot(&bqCompBunchOutQ) == DCOMP_QERR)
	{
		return;
	}


	// Take first request off queue
 
	nRequ = GrabFirstBunch(&bqCompBunchInQ);


	// If there was no request available, some other thread beat us to it

	if (nRequ == DCOMP_QEND)
	{
		UnreserveSlot(&bqCompBunchOutQ);
		return;
	}

	DispCompBunchInQ(TEXT("CPS0: "));


	// allocate return buffers

	lBunchLen = bqCompBunchInQ.abnchPool[nRequ].lBunchLen;
	lpbihUnc = &bqCompBunchInQ.abnchPool[nRequ].lpbiUnc->bmiHeader;
	lpbihRec = &bqCompBunchInQ.abnchPool[nRequ].lpbiRec->bmiHeader;

	if
	(
			((bqCompBunchInQ.abnchPool[nRequ].lplComIndex
					= midl_user_allocate(lBunchLen * sizeof(LONG))) == NULL)
		||	((bqCompBunchInQ.abnchPool[nRequ].lpadwRetFlags
					= midl_user_allocate(lBunchLen * sizeof(DWORD))) == NULL)
		||	((bqCompBunchInQ.abnchPool[nRequ].lpbComData
					= midl_user_allocate(
	// temp.: make this the full size of the uncompressed data for safety:
										 lpbihUnc->biSizeImage * lBunchLen
										 							)) == NULL)
	)
	{
		Eprintf(TEXT("CPS3: Couldn't allocate memory - will retry request\n"));
		goto RetryRequest;
	}


	// Do the actual compression

	Dprintf(TEXT("CPS4: Compressing bunch %d\n"),
									bqCompBunchInQ.abnchPool[nRequ].lBunchId);

	if (!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_IDLE))
	{
		Eprintf(TEXT("Failed to set CodecLoop thread priority below normal")
												TEXT(" before compressing\n"));
	}

	bqCompBunchInQ.abnchPool[nRequ].lReturnStatus =
	CompBunch
	(
		bqCompBunchInQ.abnchPool[nRequ].lBunchId,
		bqCompBunchInQ.abnchPool[nRequ].lBunchLen,
		bqCompBunchInQ.abnchPool[nRequ].lpbiUnc,
		bqCompBunchInQ.abnchPool[nRequ].lpbiRec,
		bqCompBunchInQ.abnchPool[nRequ].lpbUncData,
		bqCompBunchInQ.abnchPool[nRequ].lplComIndex,
		bqCompBunchInQ.abnchPool[nRequ].lpadwRetFlags,
		&bqCompBunchInQ.abnchPool[nRequ].lComDataLen,
		bqCompBunchInQ.abnchPool[nRequ].lpbComData
	);

	if (!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL))
	{
		Eprintf(TEXT("Failed to set CodecLoop thread priority normal")
												TEXT(" after compressing\n"));
	}


	// Add result to compression output queue

	EnterCriticalSection(&bqCompBunchOutQ.csQ);
	if (QueueBunch(&bqCompBunchOutQ, nRequ) == DCOMP_QERR)
	{
		Eprintf(TEXT("CPS4b: Couldn't queue recompressed bunch %ld")
				 TEXT(" for return - will retry\n"),
				 bqCompBunchInQ.abnchPool[nRequ].lBunchId);
		LeaveCriticalSection(&bqCompBunchOutQ.csQ);
		goto RetryRequest;
	}


	// Clear away request buffers no longer needed

	ClearDown(&bqCompBunchInQ.abnchPool[nRequ].lpbiUnc);
	ClearDown(&bqCompBunchInQ.abnchPool[nRequ].lpbUncData);
	LeaveCriticalSection(&bqCompBunchOutQ.csQ);

	Eprintf(
		TEXT("CPS4a: Queued recompressed bunch %d in slot %ld for return\n"),
		bqCompBunchInQ.abnchPool[nRequ].lBunchId, nRequ);


	// all OK: clear away the finished request and return

	UnreserveSlot(&bqCompBunchInQ);
	return;

	// end of normal processing



	// error processing compression request: cancel the request and re-queue
	// it on the input queue

RetryRequest:
	ClearDown(&bqCompBunchInQ.abnchPool[nRequ].lplComIndex);
	ClearDown(&bqCompBunchInQ.abnchPool[nRequ].lpadwRetFlags);
	ClearDown(&bqCompBunchInQ.abnchPool[nRequ].lpbComData);
	QueueBunchAtHead(&bqCompBunchInQ, nRequ);
	UnreserveSlot(&bqCompBunchOutQ); 
}


//
//  FUNCTION: RPCCompBunchRequest
//
//  PURPOSE:
//    place a bunch coming from the client on the input queue
//
//  PARAMETERS:
//    hHandle           - RPC binding handle - unreferenced in this function
//    lBunchId          - bunch ID (equal to first video frame number in bunch)
//    lBunchLen         - number of video frames in bunch
//    lUncFormatLen     - video format size for uncompressed data  
//    lpbUncFormat[]    - video format for uncompressed data 
//    lRecFormatLen     - video format size for recompressed data 
//    lpbRecFormat[]    - video format for recompressed data 
//    lUncDataLen       - total length of the uncompressed video data
//    lpbUncData[]      - pointer to buffer containing uncompressed video data
//
//  RETURN VALUE:
//    Non-zero on failure
//
//  COMMENTS:
//    This function is also called locally by DecompBunchPass in RECOMPRESS
//    mode.
//

LONG RPCCompBunchRequest( 
	handle_t			hHandle,
	LONG				lBunchId,
	LONG				lBunchLen,
	LONG				lUncFormatLen,
	BYTE				lpbUncFormat[],
	LONG				lRecFormatLen,
	BYTE				lpbRecFormat[],
	LONG				lUncDataLen,
	BYTE				lpbUncData[]
)
{
	LONG		nRequ;

	UNREF_PARAM(hHandle);

	Dprintf(TEXT("CBQ0: Request for bunch %d received\n"), lBunchId);


	if (gfInitPending || gfServerComputerBusy)
	{
		// waiting for queues to be cleared, or computer too busy to accept
		// RPCs - return immediately with failure code

		return(DCOMP_QERR);
	}

	if (ReserveSlot(&bqCompBunchInQ) == DCOMP_QERR)
	{
		// request queue is full - return immediately with failure code

		return(DCOMP_QERR);
	}

	if ((nRequ = GrabEmptyBunch(&bqCompBunchInQ)) == DCOMP_QEND)
	{
		// failed - return failure code

		UnreserveSlot(&bqCompBunchInQ);
		return(DCOMP_QERR);
	}

	Eprintf(TEXT("CBQ1: Request %d placed in slot %d\n"), lBunchId, nRequ);


	// Allocate buffers and copy parameter values into request structure.
	// Would be better to use persistent stub-allocated buffers without copying

	if
	(
			((bqCompBunchInQ.abnchPool[nRequ].lpbiUnc
								= midl_user_allocate(lUncFormatLen)) == NULL)
		||	((bqCompBunchInQ.abnchPool[nRequ].lpbiRec
								= midl_user_allocate(lRecFormatLen)) == NULL)
		||	((bqCompBunchInQ.abnchPool[nRequ].lpbUncData
								= midl_user_allocate(lUncDataLen)  ) == NULL)
	)
	{
		Eprintf(TEXT("CBQ3: Couldn't allocate memory for bunch %ld")
									TEXT(" - request cancelled\n"), lBunchId);
		ClearDownBunch(&bqCompBunchInQ, nRequ);
		UnreserveSlot(&bqCompBunchInQ);
		return(DCOMP_QERR);
	}

	bqCompBunchInQ.abnchPool[nRequ].lBunchId  = lBunchId;
	bqCompBunchInQ.abnchPool[nRequ].lBunchLen = lBunchLen;
	bqCompBunchInQ.abnchPool[nRequ].lUncDataLen = lUncDataLen;

	memcpy(bqCompBunchInQ.abnchPool[nRequ].lpbiUnc, lpbUncFormat,
																lUncFormatLen);
	memcpy(bqCompBunchInQ.abnchPool[nRequ].lpbiRec, lpbRecFormat,
																lRecFormatLen);
	memcpy(bqCompBunchInQ.abnchPool[nRequ].lpbUncData, lpbUncData, lUncDataLen);


	// add this request to the queue and return success code

	QueueBunch(&bqCompBunchInQ, nRequ);
	return(0);
}


//
//  FUNCTION: RPCGetCompBunchResult
//
//  PURPOSE:
//    return a compressed bunch to the client
//
//  PARAMETERS:
//    hHandle           - RPC binding handle - unreferenced in this function
//    lplReturnStatus   - pointer to var. to receive CompBunch return status
//    lBunchId          - pointer to variable containing the bunch ID of the
//                        wanted bunch.  If the ID is negative, the function
//                        sends back the first bunch on the output queue, and
//                        writes its ID into the pointed-to variable.
//    lBunchLen         - pointer to variable to receive no. of frames in bunch
//    lRecFormatLen     - video format size for recompressed data  
//    lpbRecFormat[]    - pointer to buf. to receive recompressed video format
//    lMaxBunchLen      - maximum number of frames in a bunch
//    alRecIndex[]      - pointer to buffer to receive array of indexes to the
//                        recompressed video frames in the buffer
//    dwRetFlags[]      - pointer to buffer to receive array of flags returned
//                        by the compressor for each frame
//    lMaxRecDataSize   - maximum total length of the recompressed video data
//    lplRecDataLen     - pointer to variable to receive total length of the
//                        recompressed video data
//    lpbRecData[]      - pointer to buffer to receive recompressed video data
//
//  RETURN VALUE:
//    Non-zero on failure
//
//  COMMENTS:
//

LONG RPCGetCompBunchResult( 
	handle_t			hHandle,
	LONG				*lplReturnStatus,
	LONG				*lplBunchId,
	LONG				*lplBunchLen,
	LONG				lRecFormatLen,
	BYTE				lpbRecFormat[],
	LONG				lMaxBunchLen,
	LONG				alRecIndex[],
	DWORD				dwRetFlags[],
	LONG				lMaxRecDataSize,
	LONG				*lplRecDataLen,
	BYTE				lpbRecData[]
)
{
	LONG		nRet;

	UNREF_PARAM(hHandle);
	UNREF_PARAM(lMaxRecDataSize);
	UNREF_PARAM(lMaxBunchLen);

	if (gfInitPending)
	{
		// waiting for queues to be cleared - return immediately
		// with failure code

		return(DCOMP_QERR);
	}

	if (*lplBunchId < 0)
	{
		// If requested bunch ID is negative, just take first return (if any)
		// off queue

		if ((nRet = GrabFirstBunch(&bqCompBunchOutQ)) == DCOMP_QEND)
		{
			return(DCOMP_QERR);	// no return data to send so return failure code
		}

		Eprintf(TEXT("CBR1: Returning bunch %d from slot %d\n"),
				bqCompBunchOutQ.abnchPool[nRet].lBunchId,
				nRet
			   );
	}
	else
	{
		// Otherwise find the requested bunch if present

		if ((nRet = GrabBunch(&bqCompBunchOutQ, *lplBunchId)) == DCOMP_QEND)
		{
			Dprintf(TEXT("CBR2: bunch %ld not found\n"), *lplBunchId);
			return(DCOMP_QERR);	// didn't find the bunch so return failure code
		}

		Eprintf(TEXT("CBR3: Returning bunch %d from slot %d\n"),
				bqCompBunchOutQ.abnchPool[nRet].lBunchId,
				nRet
			   );
	}


	// copy return values and buffers to output arrays

	*lplReturnStatus = bqCompBunchOutQ.abnchPool[nRet].lReturnStatus;
	*lplBunchId = bqCompBunchOutQ.abnchPool[nRet].lBunchId;
	*lplBunchLen = bqCompBunchOutQ.abnchPool[nRet].lBunchLen;
										// could check against lMaxBunchLen

	memcpy(lpbRecFormat, bqCompBunchOutQ.abnchPool[nRet].lpbiRec,
																lRecFormatLen);
	memcpy(alRecIndex, bqCompBunchOutQ.abnchPool[nRet].lplComIndex,
												*lplBunchLen * sizeof(LONG));
	memcpy(dwRetFlags, bqCompBunchOutQ.abnchPool[nRet].lpadwRetFlags,
											*lplBunchLen * sizeof(DWORD));

	*lplRecDataLen = bqCompBunchOutQ.abnchPool[nRet].lComDataLen;
									// could check against lMaxRecDataSize
	memcpy(lpbRecData, bqCompBunchOutQ.abnchPool[nRet].lpbComData,
															*lplRecDataLen);


	// clear away the bunch structure and return success code

	DispCompBunchOutQ(TEXT("CBR3: "));
	Eprintf(TEXT("Bunch %d returned after recompression\n"),
									bqCompBunchOutQ.abnchPool[nRet].lBunchId);

	ClearDownBunch(&bqCompBunchOutQ, nRet);
	UnreserveSlot(&bqCompBunchOutQ); 
	return(0);
}


//
//  FUNCTION: RPCCompBunchReqAccept
//
//  PURPOSE:
//    checks if a compression request can be accepted, i.e. the compressor's
//    input queue has space
//
//  PARAMETERS:
//    hHandle - RPC binding handle - unreferenced in this function
//
//  RETURN VALUE:
//    Non-zero if a bunch can be accepted
//
//  COMMENTS:
//

LONG RPCCompBunchReqAccept(handle_t hHandle)
{
	LONG		RetVal = (bqCompBunchInQ.nQCnt < bqCompBunchInQ.nQLen);

 	UNREF_PARAM(hHandle);

	if (gfInitPending || gfServerComputerBusy)
	{
		// waiting for queues to be cleared, or computer too busy to accept
		// RPCs - return immediately with rejection code

		return(FALSE);
	}

	Dprintf(TEXT("CAC: Request acceptance query: requests %sallowed\n"),
										(RetVal ? TEXT("") : TEXT("not ")));
	return(RetVal);
}


//
//  FUNCTION: RPCCompBunchRetAvail
//
//  PURPOSE:
//    checks to see if a specific bunch has been compressed and is available to
//    be returned
//
//  PARAMETERS:
//    hHandle  - RPC binding handle - unreferenced in this function
//    lBunchId - bunch ID of the wanted bunch.
//
//  RETURN VALUE:
//    Non-zero if the specified bunch is available.  If lBunchId is negative,
//    return non-zero if there are any completed bunches at all.
//
//  COMMENTS:
//

LONG RPCCompBunchRetAvail(handle_t hHandle, LONG lBunchId)
{
	LONG		nRet;
	LONG		RetVal = (bqCompBunchOutQ.nQHead != DCOMP_QEND);

	UNREF_PARAM(hHandle);


	if (gfInitPending)
	{
		// waiting for queues to be cleared - return immediately
		// with rejection code

		return(FALSE);
	}


	if (lBunchId < 0L)
	{
		// If requested bunch ID is negative, just check for any bunches on
		// output queue

		Dprintf(TEXT("CAV0: Availability query: %scompressed bunches available\n"),
													(RetVal ? TEXT("") : TEXT("no ")));
		return(RetVal);
	}
	else
	{
		// Otherwise search for the requested bunch

		if ((nRet = FindBunch(&bqCompBunchOutQ, lBunchId)) == DCOMP_QEND)
		{
			Dprintf(TEXT("CAV2: Availability query: bunch %ld not available\n"), lBunchId);
			return(FALSE);
		}

		Dprintf(TEXT("CAV3: Availability query: bunch %ld available\n"), lBunchId);
		return(TRUE);
	}
}


//
//  FUNCTION: InitCompQueues
//
//  PURPOSE:
//    clear down the compressor's input and output queues to the empty
//    condition
//
//  PARAMETERS:
//
//  RETURN VALUE:
//
//  COMMENTS:
//

void		InitCompQueues(void)
{
	LONG		i;

	EnterCriticalSection(&bqCompBunchInQ.csQ);
	EnterCriticalSection(&bqCompBunchOutQ.csQ);
	EnterCriticalSection(&csQPool);

	bqCompBunchInQ.nQHead = DCOMP_QEND;
	bqCompBunchInQ.nQCnt = 0;
	bqCompBunchOutQ.nQHead = DCOMP_QEND;
	bqCompBunchOutQ.nQCnt = 0;

	for (i = 0; i < BUNCHPOOLLEN; i++)
	{
		abnchQPool[i].nNext = DCOMP_QEND;
		abnchQPool[i].nStat = DCOMP_QEMPTY;

		ClearDown(&abnchQPool[i].lpbiCom);
		ClearDown(&abnchQPool[i].lpbiUnc);
		ClearDown(&abnchQPool[i].lpbiRec);
		ClearDown(&abnchQPool[i].lplComIndex);
		ClearDown(&abnchQPool[i].lpadwRetFlags);
		ClearDown(&abnchQPool[i].lpbComData);
		ClearDown(&abnchQPool[i].lpbUncData);
	}

	LeaveCriticalSection(&csQPool);
	LeaveCriticalSection(&bqCompBunchInQ.csQ);
	LeaveCriticalSection(&bqCompBunchOutQ.csQ);
}


//
//  FUNCTION: QuickDispBunch, DispCompBunchOutQ and DispCompBunchInQ
//
//  PURPOSE:
//    display routines for debugging only
//
//  PARAMETERS:
//
//  RETURN VALUE:
//
//  COMMENTS:
//

void	QuickDispBunch(LONG lBunchId, LONG lBunchLen, LONG alIndex[])
{
	LONG		i;

	Dprintf(TEXT("Bunch %ld length %ld %ld"), lBunchId, lBunchLen, alIndex[1] - alIndex[0]);
	for (i = 1; i < lBunchLen; i++)
	{
//		Dprintf(TEXT(", %ld"), alIndex[i + 1] - alIndex[i]);
//      N.B. i + 1 not valid if i == lBunchlen - 1
	}
	Dprintf(TEXT("\n"));
}

void		DispCompBunchOutQ(TCHAR *szPrefix)
{
	LONG	i, nBufPos;
	TCHAR	buf[256];

	nBufPos = _stprintf(buf, TEXT(" RET C%-2u H%-2d Q"), bqCompBunchOutQ.nQCnt, bqCompBunchOutQ.nQHead);
	
	if (bqCompBunchOutQ.nQHead != DCOMP_QEND)
	{
		nBufPos += _stprintf(&buf[nBufPos], TEXT("%d"), bqCompBunchOutQ.nQHead);

		i = bqCompBunchOutQ.nQHead;
		while (bqCompBunchOutQ.abnchPool[i].nNext != DCOMP_QEND)
		{
			i = bqCompBunchOutQ.abnchPool[i].nNext;
			nBufPos += _stprintf(&buf[nBufPos], TEXT("%d"), i);
		}
	}

	Dprintf(TEXT("%sID%08.8lx%s\n"), szPrefix, GetCurrentThreadId(), buf);
}


void		DispCompBunchInQ(TCHAR *szPrefix)
{
	LONG	i, nBufPos;
	TCHAR	buf[256];

	nBufPos = _stprintf(buf, TEXT(" REQ C%-2u H%-2d Q"), bqCompBunchInQ.nQCnt, bqCompBunchInQ.nQHead);

	if (bqCompBunchInQ.nQHead != DCOMP_QEND)
	{
		nBufPos += _stprintf(&buf[nBufPos], TEXT("%d"), bqCompBunchInQ.nQHead);

		i = bqCompBunchInQ.nQHead;
		while (bqCompBunchInQ.abnchPool[i].nNext != DCOMP_QEND)
		{
			i = bqCompBunchInQ.abnchPool[i].nNext;
			nBufPos += _stprintf(&buf[nBufPos], TEXT("%d"), i);
		}
	}
	
	Dprintf(TEXT("%sID%08.8lx%s\n"), szPrefix, GetCurrentThreadId(), buf);
}
