/* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
 * ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
 * PARTICULAR PURPOSE.
 *
 * Copyright (C) 1994  Microsoft Corporation.  All Rights Reserved.
 *
 *  MODULE:   funcs_c.c
 *
 *  PURPOSE:
 *    general-purpose functions for the distributed recompression client
 *
 *  FUNCTIONS:
 *	  midl_user_allocate(size_t len)
 *	  midl_user_free(void __RPC_FAR * ptr)
 *	  Eprintf(LPTSTR szFmt, ...)
 *	  Dprintf(LPTSTR szFmt, ...)
 *	  ErrExit(LPTSTR szFmt, ...)
 *	  GetOSPlatformId(void)
 *
 *  COMMENTS:
 *    The client application can be built to use either UNICODE or ANSI APIs,
 *    according to the compilation settings
 */

#include	"dcomp.h"
#include	"dcomp_c.h"


//
//  FUNCTION: midl_user_allocate
//
//  PURPOSE: memory allocation routine for the application 
//
//  PARAMETERS:
//    len - size of memory block to allocate, in bytes
//
//  RETURN VALUE:
//    Pointer to the allocated space, or NULL if the requested memory could not
//    be allocated
//
//  COMMENTS:
//    Required by RPC
//

void __RPC_FAR * __RPC_USER midl_user_allocate(size_t len)
{
	void __RPC_FAR * __RPC_USER ptr = malloc(len);

	if (ptr == NULL)
	{
		Eprintf(TEXT("Client couldn't allocate %u bytes\n"), len, ptr);
	}
	else
	{
		Dprintf(TEXT("Client allocated %u bytes at 0x%08.8lx\n"), len, ptr);
	}
	return(ptr);
}


//
//  FUNCTION: midl_user_free
//
//  PURPOSE: memory free routine for the application 
//
//  PARAMETERS:
//    ptr - pointer to the memory block to be freed
//
//  RETURN VALUE:
//
//  COMMENTS:
//    Required by RPC
//

void __RPC_USER midl_user_free(void __RPC_FAR * ptr)
{
	free(ptr);
	Dprintf(TEXT("Client freed memory at 0x%08.8lx\n"), ptr);
}


//
//  FUNCTION: Eprintf
//
//  PURPOSE: formatted print to standard error
//
//  PARAMETERS:
//    As for printf
//
//  RETURN VALUE:
//    As for printf
//
//  COMMENTS:
//

int			Eprintf(LPTSTR szFmt, ...)
{
	int			retval;
	va_list		vl;

	va_start(vl, szFmt);
	retval = _vftprintf(stderr, szFmt, vl);
	va_end(vl);
	return(retval);
}


//
//  FUNCTION: Dprintf
//
//  PURPOSE: debug formatted print to standard error
//
//  PARAMETERS:
//    As for printf
//
//  RETURN VALUE:
//    As for printf
//
//  COMMENTS:
//    Only prints if global flag gfDebug is TRUE (non-zero)
//

int			Dprintf(LPTSTR szFmt, ...)
{
	int			retval = 0;
	va_list		vl;

	if (gfDebug)
	{
		va_start(vl, szFmt);
		retval = _vftprintf(stderr, szFmt, vl);
		va_end(vl);
	}
	return(retval);
}


//
//  FUNCTION: ErrExit
//
//  PURPOSE: terminate execution after printing message to standard error
//
//  PARAMETERS:
//    As for printf
//
//  RETURN VALUE:
//    As for printf
//
//  COMMENTS:
//

void		ErrExit(LPTSTR szFmt, ...)
{
	va_list		vl;

	va_start(vl, szFmt);
	_vftprintf(stderr, szFmt, vl);
	va_end(vl);
	exit(1);
}


//
//  FUNCTION: GetOSPlatformId
//
//  PURPOSE: Get current operating system platform ID from system
//
//  PARAMETERS:
//
//  RETURN VALUE:
//    Platform ID if successful
//	  	  -1      if unsuccessful
//
//  COMMENTS:
//	  Platform IDs defined include:
//
//		Value						Platform
//
//		VER_PLATFORM_WIN32s			Win32s on Windows 3.1 
//		VER_PLATFORM_WIN32_WINDOWS	Win32 on Chicago
//		VER_PLATFORM_WIN32_NT		Windows NT 
//

DWORD		GetOSPlatformId(void)
{
	OSVERSIONINFO	osversioninfo;

	osversioninfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if (!GetVersionEx(&osversioninfo))
	{
		Eprintf(TEXT("Couldn't get operating system version info\n"));
		return((DWORD) -1);
	}

	Dprintf(TEXT("Operating system platform ID is %lu\n"),
												osversioninfo.dwPlatformId);

	return(osversioninfo.dwPlatformId);
}
